# -*- coding: utf-8 -*-
# Copyright (c) 2011 University of Jyväskylä and Contributors.
#
# All Rights Reserved.
#
# Authors:
#     Esa-Matti Suuronen <esa-matti@suuronen.org>
#     Asko Soukka <asko.soukka@iki.fi>
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS

import time

from ZServer.HTTPServer import zhttp_server
from App.config import getConfiguration

from sauna.reload import\
    autoinclude, forkloop, reload_paths, watcher, monkeypatcher
from sauna.reload.utils import errline, logger


def startForkLoop(event):

    if not monkeypatcher.PATCHED:
        errline()
        errline("sauna.reload is not installed correctly!")
        errline("Your are missing following line from instance part "
                "of your buildout:")
        errline()
        errline("    zope-conf-additional = %import sauna.reload")
        errline()
        errline("Not starting fork loop.")
        errline()
        return

    if not reload_paths:
        errline()
        errline("sauna.reload: No paths in RELOAD_PATH environment variable. "
                "Not starting fork loop.")
        errline("Set it to your development egg paths to activate reloading")
        errline()
        errline("Example:")
        errline("         $ RELOAD_PATH=src/ bin/instance fg")
        errline()
        return

    # Start fs monitor before the forkloop
    watcher.Watcher(reload_paths.getParentPaths(), forkloop).start()

    # Build and execute a configuration file to include meta, configuration and
    # overrides for dependencies of the deferred development packages.
    autoinclude.includeDependenciesForDeferred()
    autoinclude.checkDeferringErrors()

    config = getConfiguration()
    zserver = [server for server in config.servers
        if isinstance(server, zhttp_server)][0]

    logger.info("We saved at least %s seconds from boot up time" %
        (time.time() - forkloop.boot_started))

    logger.info("Overview available at http://127.0.0.1:%i/@@saunareload" %
        (zserver.port))

    forkloop.start()



# -*- coding: utf-8 -*-
# Copyright (c) 2011 University of Jyväskylä and Contributors.
#
# All Rights Reserved.
#
# Authors:
#     Esa-Matti Suuronen <esa-matti@suuronen.org>
#     Asko Soukka <asko.soukka@iki.fi>
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS

"""Database Adapters"""

from persistent.TimeStamp import TimeStamp

from zope.interface import implements
from zope.component import adapts

from ZODB.interfaces import IDatabase
from ZODB.FileStorage.FileStorage import FileStorage, read_index
from ZEO.ClientStorage import ClientStorage

from sauna.reload.interfaces import IDatabaseHooks


class ZODBDatabaseHooksAdapter(object):
    """
    Selective ZODB-adapter (e.g. supporting both FileStorage and ZEO)
    """
    implements(IDatabaseHooks)
    adapts(IDatabase)

    def __init__(self, context):
        self.context = IDatabaseHooks(context.storage)

    def prepareForReload(self):
        return self.context.prepareForReload()

    def resumeFromReload(self):
        return self.context.resumeFromReload()


class ZODBFileStorageDatabaseHooksAdapter(object):
    """
    FileStorage-adapter
    """
    implements(IDatabaseHooks)
    adapts(FileStorage)

    def __init__(self, context):
        # Try to get the *real* FileStorage,
        # because `context` may be just a BlobStorage-wrapper
        # and it wraps FileStorage differently between
        # ZODB3-3.9.5 and 3.10.x-series (eg. between Plone 4.0 and 4.1).
        self.context = getattr(context, '_BlobStorage__storage', context)

    def prepareForReload(self):
        # Save ``Data.fs.index`` before dying
        # to notify the next child
        # about all the presistent changes
        self.context._lock_acquire()
        try:
            self.context._save_index()
        finally:
            self.context._lock_release()

    def resumeFromReload(self):
        # Reload ``Data.fs.index``
        # to get up to date
        # about all the persistent changes
        self.context._lock_acquire()
        try:
            # Get indexes in their pre-fork state
            index, tindex = (self.context._index, self.context._tindex)
            # Load saved ``Data.fs.index`` to see the persistent changes
            # created by the previous child.
            index, start, ltid =\
                self.context._restore_index() or (None, None, None)
            # Sanity check. Last transaction in restored index must match
            # the last transaction given by FileStorage transaction iterator.
            if ltid and ltid == tuple(self.context.iterator())[-1].tid:
                self.context._initIndex(index, tindex)
                self.context._pos, self.context._oid, tid = read_index(
                    self.context._file, self.context._file_name, index, tindex,
                    stop='\377' * 8, ltid=ltid, start=start, read_only=False)
                self.context._ltid = tid
                self.context._ts = TimeStamp(tid)
        finally:
            self.context._lock_release()


class ZEOClientStorageDatabaseHooksAdapter(object):
    """
    ZEO-client adapter
    """
    implements(IDatabaseHooks)
    adapts(ClientStorage)

    def __init__(self, context):
        self.context = context

        # Store cache settings
        self._cache_path = self.context._cache.path
        self._cache_size = self.context._cache.maxsize

        # Close the main process' connection (before forkloop)
        _rpc_mgr = self.context._rpc_mgr
        _rpc_mgr.close()

        # Close the main process' cache (before forkloop)
        self.context._cache.close()

    def prepareForReload(self):
        _rpc_mgr = self.context._rpc_mgr
        if self.context.is_connected():
            # Close the connection (before the child is killed)
            _rpc_mgr.close()

            # Close the cache (before the child is killed)
            self.context._cache.close()

    def resumeFromReload(self):
        # Open a new cache for the new child
        self.context._cache = self.context.ClientCacheClass(
            self._cache_path, size=self._cache_size)

        # Prepare a new connection for the new child
        self.context._rpc_mgr = self.context.ConnectionManagerClass(
            self.context._addr, self.context, tmin=1, tmax=30)

        # Connect the new child to ZEO
        self.context._rpc_mgr.attempt_connect()
        if not self.context._rpc_mgr.attempt_connect():
            self.context._rpc_mgr.connect()



# -*- coding: utf-8 -*-
"""Configures p.a.theming-plugins for *sauna.reload*-controlled products"""

from plone.resource.utils import iterDirectoriesOfType

from plone.app.theming.interfaces import THEME_RESOURCE_NAME

from plone.app.theming.plugins.utils import getPlugins
from plone.app.theming.plugins.utils import getPluginSettings

from sauna.reload import reload_paths


def onStartup(event):
    """
    Call onDiscovery() on each plugin for each theme on startup
    """

    plugins = getPlugins()

    for themeDirectory in iterDirectoriesOfType(THEME_RESOURCE_NAME):
        if themeDirectory.directory in reload_paths:  # only for sauna.reload!
            pluginSettings = getPluginSettings(themeDirectory, plugins)

            for name, plugin in plugins:
                plugin.onDiscovery(themeDirectory.__name__,
                                   pluginSettings[name],
                                   pluginSettings)



# -*- coding: utf-8 -*-
# Copyright (c) 2011 University of Jyväskylä and Contributors.
#
# All Rights Reserved.
#
# Authors:
#     Esa-Matti Suuronen <esa-matti@suuronen.org>
#     Asko Soukka <asko.soukka@iki.fi>
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS

"""The Fork Loop (tm)"""

import time
import os
import signal
import atexit

from zope.event import notify
from App.config import getConfiguration

from Signals.SignalHandler import SignalHandler
registerHandler = SignalHandler.registerHandler

from sauna.reload import autoinclude, fiveconfigure
from sauna.reload.interfaces import IDatabaseHooks
from sauna.reload.events import NewChildForked, NewChildIsReady
from sauna.reload.utils import errline, logger


class CannotSpawnNewChild(Exception):
    pass


class ForkLoop(object):

    def __init__(self):

        self.fork = True  # Create child on start
        self.active = False
        self.pause = False
        self.killed_child = True
        self.forking = False
        self.exit = False

        self.parent_pid = os.getpid()
        self.child_pid = None

        # Timers
        self.boot_started = None
        self.child_started = None

        self.cfg = None
        self.database = None

    def isChild(self):
        return self.child_pid == 0

    def startBootTimer(self):
        if not self.boot_started:
            self.boot_started = time.time()

    def startChildBooTimer(self):
        self.child_started = time.time()

    def isChildAlive(self):

        if self.isChild():
            return True

        return (self.child_pid is not None
            and os.path.exists("/proc/%i" % self.child_pid))

    def _scheduleFork(self, signum=None, frame=None):
        self.fork = True

    def _childIsGoingToDie(self, signum=None, frame=None):
        self.killed_child = True

    def start(self):
        """
        Start fork loop.
        """

        # Load configuration here to make sure that everything for it is loaded
        self.cfg = getConfiguration()

        # Must import here because we don't have DB on bootup yet
        from Globals import DB
        self.database = IDatabaseHooks(DB)

        # SIGCHLD tells us that child process has really died and we can spawn
        # new child
        registerHandler(signal.SIGCHLD, self._waitChildToDieAndScheduleNew)

        # With SIGUSR1 child can tell that it dies by request, not by exception
        # etc.
        registerHandler(signal.SIGUSR1, self._childIsGoingToDie)

        self.loop()

    def loop(self):
        """
        Magic happens here
        """

        registerHandler(signal.SIGINT, self._parentExitHandler)
        registerHandler(signal.SIGTERM, self._parentExitHandler)

        self.active = True

        logger.info("Fork loop starting on parent. PID %i" % os.getpid())
        while True:
            self.forking = False

            if self.exit:
                return

            if self.fork:
                self.fork = False

                if self.pause:
                    # Pause mode. No forks now.
                    continue

                if not self.killed_child:
                    errline()
                    errline("Child died on bootup. "
                            "Pausing fork loop for now.")
                    errline("Fix possible errors and save edits "
                            "and we'll try booting again.")
                    errline("Waiting...")

                    # Child died because of unknown reason. Mark it as killed
                    # and go into pause mode.
                    self.killed_child = True
                    self.pause = True
                    continue

                if self.isChildAlive():
                    # Child is still alive for some reason. Lets wait few
                    # rounds for it to die.
                    continue

                self.forking = True
                self.startChildBooTimer()
                self.child_pid = os.fork()
                if self.child_pid == 0:
                    break
                self.killed_child = False

            time.sleep(1)

        logger.setChildLogger()
        logger.info("Forked new child. Installing reloadable products...")

        self._prepareNewChild()

        self.forking = False

        logger.logDeferredErrors()
        logger.logDeferredWarnings()
        logger.logDeferred()

        logger.info("Booted up new child in %s seconds. PID %i" % (
            time.time() - self.child_started,  os.getpid()))

        notify(NewChildIsReady(self))

    def _prepareNewChild(self):
        """
        Prepare newly forked child. Make sure that it can properly read DB
        and install deferred products.
        """

        # Register exit listener. We cannot immediately spawn new child when we
        # get a modified event. Must wait that child has closed database etc.
        atexit.register(self._childExitHandler)

        # Make sure that PID files and locks stay here, because dying child
        # will clear them.
        self.makeLockFile()
        self.makePidFile()

        self.database.resumeFromReload()

        notify(NewChildForked(self))

        autoinclude.includeDeferred()
        fiveconfigure.installDeferred()

    def spawnNewChild(self):
        """
        STEP 1 (parent): New child spawning starts by killing the current
        child.
        """

        if not self.active:
            raise CannotSpawnNewChild("Loop not started yet")

        if self.forking:
            raise CannotSpawnNewChild("Serious forking action is already "
                                      "going on. Cannot fork now.")

        if self.child_pid is None:
            raise CannotSpawnNewChild("No killing yet. Not started child yet")

        self.pause = False

        if not self.killed_child or self.isChild():
            self._killChild()
        else:
            # Ok, we already have sent the SIGINT the child, but asking for new
            # child
            logger.info("Not sending SIGINT because we already killed "
                        "the child. Just scheduling new fork.")
            self._scheduleFork()

        self.killed_child = True

    def _killChild(self):
        if self.isChild():
            # Signal parent that this is requested kill, not an error situation
            os.kill(self.parent_pid, signal.SIGUSR1)
            # Kill itself
            os.kill(os.getpid(), signal.SIGINT)
        else:
            os.kill(self.child_pid, signal.SIGINT)

    def _parentExitHandler(self, signum=None, frame=None):
        if self.isChild():
            return

        self.exit = True

        if self.isChildAlive():
            logger.info("Parent dying. Killing child first.")
            self._killChild()

    def _childExitHandler(self):
        """
        STEP 2 (child): Child is about to die. Fix DB.
        """

        self.database.prepareForReload()

    def _resolveExitReason(self, exit_status):
        """
        Resolve the reason for (abnormal) child termination

        :param exit_status: Whatever os.wait() pukes on you
        """

        flags = []

        if os.WCOREDUMP(exit_status):
            flags.append("core dumped")

        if os.WIFSIGNALED(exit_status):
            code = os.WTERMSIG(exit_status)
            flags.append("terminated by signal %d" % code)

        if os.WIFEXITED(exit_status):
            code = os.WEXITSTATUS(exit_status)
            flags.append("exited with code %d" % code)

        return ", ".join(flags)

    def _waitChildToDieAndScheduleNew(self, signal=None, frame=None):
        """
        STEP 3 (parent): Child told us via SIGCHLD that we can spawn new child
        """
        try:
            # Acknowledge dead child
            pid, exit_status = os.wait()
            logger.info("Child %d exited, reasons: %s" % (
                pid, self._resolveExitReason(exit_status)))
        except OSError:
            # OSError: [Errno 10] No child processes
            pass
        # Schedule new
        self._scheduleFork()

    # Modified from Zope2/Startup/__init__.py
    def makePidFile(self):

        if not self.cfg.zserver_read_only_mode:
            # write the pid into the pidfile if possible
            try:
                if os.path.exists(self.cfg.pid_filename):
                    os.unlink(self.cfg.pid_filename)
                f = open(self.cfg.pid_filename, 'w')
                f.write(str(self.parent_pid))
                f.close()
            except IOError:
                pass

    # Modified from Zope2/Startup/__init__.py
    def makeLockFile(self):
        if not self.cfg.zserver_read_only_mode:
            from Zope2.Startup.misc.lock_file import lock_file
            lock_filename = self.cfg.lock_filename
            try:
                if os.path.exists(lock_filename):
                    os.unlink(lock_filename)
                lockfile = open(lock_filename, 'w')
                lock_file(lockfile)
                lockfile.write(str(self.parent_pid))
                lockfile.flush()
            except IOError:
                pass



# -*- coding: utf-8 -*-
# Copyright (c) 2011 University of Jyväskylä and Contributors.
#
# All Rights Reserved.
#
# Authors:
#     Esa-Matti Suuronen <esa-matti@suuronen.org>
#     Asko Soukka <asko.soukka@iki.fi>
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS

"""ZCA Interface Definitions"""

from zope.interface import Interface, Attribute


class IDatabaseHooks(Interface):
    """
    Provides storage-specific hooks to be called during the reload.
    """

    def prepareForReload():
        """
        Is called before the reload (before the process is killed)
        to allow database connection be prepared for it.
        """

    def resumeFromReload():
        """
        Is called after the reload (after a new process has been spawned)
        to allow database connection be restored.
        """


class INewChildForked(Interface):
    """
    Emited immediately after new process is forked. No development packages
    have been yet installed.

    Useful if you want to do something before your code gets loaded.

    Note that you cannot listen this event on a package that is marked for
    reloading as it is not yet installed when this is fired.
    """

    forkloop = Attribute('ForkLoop instance')


class INewChildIsReady(Interface):
    """
    Emitted when all the development packages has been installed to the new
    forked child.

    Useful for notifications etc.
    """

    forkloop = Attribute('ForkLoop instance')



# -*- coding: utf-8 -*-
# Copyright (c) 2011 University of Jyväskylä and Contributors.
#
# All Rights Reserved.
#
# Authors:
#     Esa-Matti Suuronen <esa-matti@suuronen.org>
#     Asko Soukka <asko.soukka@iki.fi>
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.

"""Module implementation loader with Zope monkey patching abilities"""

import imp
import os
from pkgutil import ImpLoader

from sauna.reload import autoinclude, fiveconfigure

PATCHED = False


class MonkeyPatchingLoader(ImpLoader):
    """
    Lucky for us ZConfig will use PEP 302 module hooks to load this file,
    and this class implements a get_data hook to intercept the component.xml
    loading and give us a point to generate it.

    Thanks to Martijn Pieters for teaching us this trick.
    """

    def __init__(self, module):
        name = module.__name__
        path = os.path.dirname(module.__file__)
        description = ("", "", imp.PKG_DIRECTORY)
        ImpLoader.__init__(self, name, None, path, description)

    def get_data(self, pathname):
        global PATCHED
        PATCHED = True
        if os.path.split(pathname) == (self.filename, 'component.xml'):
            from sauna.reload import reload_paths
            if reload_paths:
                # 1) Defer autoinclude of packages found under reload paths.
                autoinclude.deferConfigurations()
                # 2) Prevent Five from finding packages under reload paths.
                fiveconfigure.deferInstalls()
            # 3) Return dummy config to keep Zope happy.
            return '<component></component>'
        return super(MonkeyPatchingLoader, self).get_data(self, pathname)



# -*- coding: utf-8 -*-
# Copyright (c) 2011 University of Jyväskylä and Contributors.
#
# All Rights Reserved.
#
# Authors:
#     Asko Soukka <asko.soukka@iki.fi>
#     Esa-Matti Suuronen <esa-matti@suuronen.org>
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS

"""Utilities for deferring product installations under selected paths"""

from Zope2.App.zcml import load_config


def is_sauna_product(product_dir):
    """
    Check if a product dir is reloadable.

    :param product_dir: Absolute path to an egg
    """
    from sauna.reload import reload_paths

    if reload_paths.hasAbsPath(product_dir):
        return True

    return False


# Monkey-patched to delay sauna.reloadable Products.xxx eggs loads
# by Zope 2.13.
def install_products(app):
    # Install a list of products into the basic folder class, so
    # that all folders know about top-level objects, aka products
    #
    from OFS.Application import get_folder_permissions
    from OFS.Application import install_product
    from OFS.Application import install_package
    from OFS.Application import get_packages_to_initialize
    from OFS.Application import get_products

    from App.config import getConfiguration
    import transaction

    folder_permissions = get_folder_permissions()
    meta_types = []
    done = {}

    debug_mode = getConfiguration().debug_mode

    transaction.get().note('Prior to product installs')
    transaction.commit()

    products = get_products()

    for priority, product_name, index, product_dir in products:
        # For each product, we will import it and try to call the
        # intialize() method in the product __init__ module. If
        # the method doesnt exist, we put the old-style information
        # together and do a default initialization.

        if product_name in done:
            continue

        if is_sauna_product(product_dir):
            # Will be later loaded by installDeferred()
            continue

        done[product_name] = 1
        install_product(app, product_dir, product_name, meta_types,
                        folder_permissions, raise_exc=debug_mode)

    # Delayed install of packages-as-products
    for module, init_func in tuple(get_packages_to_initialize()):
        install_package(app, module, init_func, raise_exc=debug_mode)


def findProducts():
    """
    Do not find products under sauna.reload's reload paths.
    """
    import Products
    from types import ModuleType
    from sauna.reload import reload_paths
    products = []
    for name in dir(Products):
        product = getattr(Products, name)
        if isinstance(product, ModuleType) and hasattr(product, '__file__'):
            # Do not find products under sauna.reload's reload paths
            if not getattr(product, '__file__') in reload_paths:
                products.append(product)
    return products


def findDeferredProducts():
    """
    Find only those products, which are under some reload path.
    """
    import Products
    from types import ModuleType
    from sauna.reload import reload_paths
    products = []
    for name in dir(Products):
        product = getattr(Products, name)
        if isinstance(product, ModuleType) and hasattr(product, '__file__'):
            # Find only products under sauna.reload's reload paths
            if getattr(product, '__file__') in reload_paths:
                products.append(product)
    return products


def deferInstalls():
    """
    Patch fiveconfigure with findProducts-method,
    which is unable to see the reloaded paths.
    """
    try:
        import OFS.metaconfigure
        import OFS.Application
        setattr(OFS.metaconfigure, 'findProducts', findProducts)
        setattr(OFS.Application, 'install_products', install_products)

    except ImportError:
        import Products.Five.fiveconfigure
        setattr(Products.Five.fiveconfigure, 'findProducts', findProducts)


def installDeferred():
    """
    Temporarily patch fiveconfigure with a findProducts-method,
    which is able to see only the products under the reload paths
    and execute Five configuration directives for those.
    """
    import sauna.reload
    try:
        # Zope 2.13
        import OFS.metaconfigure
        setattr(OFS.metaconfigure, 'findProducts', findDeferredProducts)
        load_config('fiveconfigure.zcml', sauna.reload)
        setattr(OFS.metaconfigure, 'findProducts', findProducts)
    except ImportError:
        # Zope 2.12
        import Products.Five.fiveconfigure
        setattr(Products.Five.fiveconfigure, 'findProducts',
                findDeferredProducts)
        load_config('fiveconfigure.zcml', sauna.reload)
        setattr(Products.Five.fiveconfigure, 'findProducts', findProducts)

    # Five pushes old-style product initializations into
    # Products._packages_to_initialize-list.
    # We must loop through that list to find the reloaded packages
    # and try to install them when found.
    from App.config import getConfiguration
    from OFS.Application import install_package
    from Zope2.App.startup import app
    app = app()
    debug_mode = getConfiguration().debug_mode
    from sauna.reload import reload_paths
    try:
        # Zope 2.13
        import OFS.metaconfigure
        # We iterate a copy of packages_to_initialize,
        # because install_package mutates the original.
        packages_to_initialize = [info for info in getattr(
            OFS.metaconfigure, '_packages_to_initialize', [])]
        for module, init_func in packages_to_initialize:
            if getattr(module, '__file__') in reload_paths:
                install_package(app, module, init_func, raise_exc=debug_mode)
    except ImportError:
        # Zope 2.12
        import Products
        # We iterate a copy of packages_to_initialize,
        # because install_package mutates the original.
        packages_to_initialize = [info for info in getattr(
            Products, '_packages_to_initialize', [])]
        for module, init_func in packages_to_initialize:
            if getattr(module, '__file__') in reload_paths:
                install_package(app, module, init_func, raise_exc=debug_mode)
        if hasattr(Products, '_packages_to_initialize'):
            del Products._packages_to_initialize



# -*- coding: utf-8 -*-
# Copyright (c) 2011 University of Jyväskylä and Contributors.
#
# All Rights Reserved.
#
# Authors:
#     Esa-Matti Suuronen <esa-matti@suuronen.org>
#     Asko Soukka <asko.soukka@iki.fi>
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.

"""Watchdog-plugin to trigger reload when the developed files are modified"""

import signal

from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler

from Signals.SignalHandler import SignalHandler
registerHandler = SignalHandler.registerHandler

from sauna.reload.forkloop import CannotSpawnNewChild
from sauna.reload.utils import logger


class Watcher(FileSystemEventHandler):

    allowed_extensions = set(("py", "zcml", "po"))

    def __init__(self, paths, forkloop):
        self.forkloop = forkloop
        self.paths = paths
        FileSystemEventHandler.__init__(self)
        self.observers = []

    def start(self):
        """
        Start file monitoring thread
        """

        registerHandler(signal.SIGINT, self._exitHandler)
        registerHandler(signal.SIGTERM, self._exitHandler)

        for path in self.paths:
            logger.info("Starting file monitor on %s" % path)
            observer = Observer()
            self.observers.append(observer)
            observer.schedule(self, path=path, recursive=True)
            try:
                observer.start()
            except KeyError:
                pass

    def _exitHandler(self):
        for obs in self.observers:
            try:
                obs.stop()
            except KeyError:
                pass

    def on_any_event(self, event):
        ext = event.src_path.split(".")[-1].lower()
        if ext not in self.allowed_extensions:
            return

        logger.info("Got '%s' event on %s" %
            (event.event_type, event.src_path))

        try:
            self.forkloop.spawnNewChild()
        except CannotSpawnNewChild as e:
            logger.error(str(e.args[0]))



# -*- coding: utf-8 -*-
# Copyright (c) 2011 University of Jyväskylä and Contributors.
#
# All Rights Reserved.
#
# Authors:
#     Asko Soukka <asko.soukka@iki.fi>
#     Esa-Matti Suuronen <esa-matti@suuronen.org>
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS

"""Utilities for deferring autoinclude configurations under selected paths"""

from pkg_resources import iter_entry_points

from z3c.autoinclude.dependency import DependencyFinder
from Zope2.App.zcml import load_config, load_string

import os
from pkg_resources import get_provider
from pkg_resources import working_set as ws
from zope.dottedname.resolve import resolve
from z3c.autoinclude.utils import DistributionManager
from z3c.autoinclude.utils import ZCMLInfo

DEFERRED_TARGET = 'sauna.reload'
FAILED_TO_DEFER = []


def deferConfigurations():
    """
    Defer autoincluding of configurations
    for the packages under the given paths
    by changing *z3c.autoinclude.plugin* -entrypoints
    to target *sauna.reload* instead of *plone*.
    """
    from sauna.reload import reload_paths

    for ep in iter_entry_points('z3c.autoinclude.plugin'):
        if ep.dist.location in reload_paths\
            and ep.dist.project_name != 'sauna.reload':
            ep.module_name = DEFERRED_TARGET


def getDependencyInfosForDeferred():
    """
    Return dictionary with lists of configuration files
    for the dependencies of the deferred eggs.
    """
    deferred = ['zc.table', 'hurry.workflow']
    # XXX: Assuming that all dependencies should be autoincluded
    # will probably get us into trouble, but let's see how big trouble.
    # *zc.table* is an example of a dependency, whose *configure.zcml*
    # will not run in Plone environment.
    # Some better solution than just a blacklist would be welcome.
    from sauna.reload import reload_paths
    zcml_to_look_for = ('meta.zcml', 'configure.zcml', 'overrides.zcml')
    deps = ZCMLInfo(zcml_to_look_for)
    for ep in iter_entry_points('z3c.autoinclude.plugin'):
        if ep.module_name == DEFERRED_TARGET:
            deferred.append(ep.dist.project_name)
            # XXX: We cannot call DependencyFinder(ep.dist).includableInfo,
            # because it eventually imports also our development packages
            # while resolving existence of *zcml_to_look_for*.
            finder = DependencyFinder(ep.dist)
            info = ZCMLInfo(zcml_to_look_for)
            for req in finder.context.requires():
                # Skip missing and deferred requirements
                dist = ws.find(req)  # find req from the current working set
                if dist is None or dist.location in reload_paths:
                    continue
                # Resolve ZCMLs to be loaded for the other requirements
                dist_manager = DistributionManager(get_provider(req))
                for dotted_name in dist_manager.dottedNames():
                    module = resolve(dotted_name)
                    for candidate in zcml_to_look_for:
                        candidate_path = os.path.join(
                            os.path.dirname(module.__file__), candidate)
                        if os.path.isfile(candidate_path):
                            info[candidate].append(dotted_name)
            for key in deps:
                deps[key].extend(info.get(key, []))
    for key in deps:
        deps[key] = set([n for n in deps[key] if n not in deferred])
    return deps


def includeDependenciesForDeferred():
    """
    Build and execute a configuration file
    to include meta, configuration and overrides
    for dependencies of the deferred development packages.
    """
    deps = getDependencyInfosForDeferred()
    config = u"""\
<configure
    xmlns="http://namespaces.zope.org/zope"
    xmlns:zcml="http://namespaces.zope.org/zcml">

    """ + u"".join([u"""<include
        zcml:condition="not-have disable-autoinclude"
        package="%s"
        file="meta.zcml"
        />""" % name for name in deps.get('meta.zcml', ())]) + """

    """ + u"".join([u"""<include
        zcml:condition="not-have disable-autoinclude"
        package="%s"
        file="configure.zcml"
        />""" % name for name in deps.get('configure.zcml', ())]) + """

    """ + u"".join([u"""<includeOverrides
        zcml:condition="not-have disable-autoinclude"
        package="%s"
        file="overrides.zcml"
        />""" % name for name in deps.get('overrides.zcml', ())]) + """

</configure>"""
    load_string(config)


def checkDeferringErrors():
    """
    Check if the configuration files were deferred successfully.
    """
    try:
        from Zope2.App.zcml import _context as configuration_context
        configuration_context  # pyflakes
    except ImportError:
        from Products.Five.zcml import _context as configuration_context

    from sauna.reload import reload_paths
    from sauna.reload.utils import logger

    FAILED_TO_DEFER = []
    cwd = os.getcwd() + os.path.sep
    for zcml in getattr(configuration_context, '_seen_files', ()):
        if zcml in reload_paths:
            logger.deferredWarning("Cannot reload %s." % zcml.replace(cwd, ''))
            FAILED_TO_DEFER.append(zcml)


def includeDeferred():
    """
    Autoinclude configuration files for the deferred packages
    and log the successfully reloaded configuration files.
    """
    import sauna.reload
    load_config('autoinclude.zcml', sauna.reload)

    # Log the results
    try:
        from Zope2.App.zcml import _context as configuration_context
        configuration_context  # pyflakes
    except ImportError:
        from Products.Five.zcml import _context as configuration_context

    from sauna.reload import reload_paths
    from sauna.reload.utils import logger

    cwd = os.getcwd() + os.path.sep
    for zcml in getattr(configuration_context, '_seen_files', ()):
        if zcml in reload_paths and zcml not in FAILED_TO_DEFER:
            if zcml.startswith(os.path.dirname(sauna.reload.__file__)):
                continue
            logger.deferred("Reloaded %s." % zcml.replace(cwd, ''))



# -*- coding: utf-8 -*-
# Copyright (c) 2011 University of Jyväskylä and Contributors.
#
# All Rights Reserved.
#
# Authors:
#     Esa-Matti Suuronen <esa-matti@suuronen.org>
#     Asko Soukka <asko.soukka@iki.fi>
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.

"""Logging utilities"""

import sys
import logging


class LoggerWrapper(object):

    def __init__(self, name):
        self.name = name
        self._deferred = []
        self._deferredWarnings = []
        self._deferredErrors = []
        self.setParentLogger()

    def deferred(self, msg):
        self._deferred.append(msg)

    def logDeferred(self):
        for msg in self._deferred:
            self.info(msg)
        self._deferred = []

    def deferredError(self, msg):
        self._deferredErrors.append(msg)

    def logDeferredErrors(self):
        for msg in self._deferredErrors:
            self.error(msg)
        self._deferredErrors = []

    def deferredWarning(self, msg):
        self._deferredWarnings.append(msg)

    def logDeferredWarnings(self):
        for msg in self._deferredWarnings:
            self.warn(msg)
        self._deferredWarnings = []

    def setParentLogger(self):
        suffix = '.parent'
        self.logger = logging.getLogger(self.name + suffix)

    def setChildLogger(self):
        suffix = '.child'
        self.logger = logging.getLogger(self.name + suffix)

    def __getattr__(self, name):
        return getattr(self.logger, name)


logger = LoggerWrapper('sauna.reload')


def errline(msg="", *rest):
    rest = ' '.join(map(str, rest))
    sys.stderr.write(str(msg) + ' ' + rest + '\n')



# -*- coding: utf-8 -*-
# Copyright (c) 2011 University of Jyväskylä and Contributors.
#
# All Rights Reserved.
#
# Authors:
#     Esa-Matti Suuronen <esa-matti@suuronen.org>
#     Asko Soukka <asko.soukka@iki.fi>
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.

"""ReloadPaths-utility to ease reload path related operations"""

import os


class ReloadPaths(object):

    def __init__(self, paths):
        self.paths = [p.rstrip(os.sep) for p in paths]

    def __nonzero__(self):
        return len(self.paths) > 0

    def __contains__(self, test_path):
        test_path = test_path.rstrip(os.sep)

        for path in self.paths:
            if test_path.startswith(path):
                return True
        return False

    def __iter__(self):
        return list(self.paths)

    def findEggPaths(self):

        # TODO: could read directly from configure objects.
        egg_paths = []

        for path in self.paths:
            if os.path.exists(os.path.join(path, 'setup.py')):
                egg_paths.append(path)
            else:
                for dirpath, dirnames, filenames in os.walk(path):
                    if 'setup.py' in filenames:
                        egg_paths.append(dirpath)

        return sorted(egg_paths)

    def getParentPaths(self):
        parents = []

        parent = None

        for path in sorted(self.paths):

            if parent is None:
                parent = path
                continue

            if path.startswith(parent):
                continue
            else:
                parents.append(parent)
                parent = path

        if parent not in parents:
            parents.append(parent)

        return parents

    def hasAbsPath(self, test_path):
        """
        Check if we have absolute folder path of an egg on the reload list.
        """
        test_path = os.path.abspath(test_path)
        for path in self.paths:
            ap = os.path.abspath(path)
            if test_path.startswith(ap):
                return True

        return False


if __name__ == '__main__':
    paths = [
        '/foo/bar',
        '/newparent',
        '/foo',
        '/foo/child',
        '/another/one',
    ]
    rp = ReloadPaths(paths)
    print list(rp.getParentPaths())



# -*- coding: utf-8 -*-
# Copyright (c) 2011 University of Jyväskylä and Contributors.
#
# All Rights Reserved.
#
# Authors:
#     Esa-Matti Suuronen <esa-matti@suuronen.org>
#     Asko Soukka <asko.soukka@iki.fi>
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.

"""
sauna.reload
============

Enable sauna.reload's Zope patches and deferrend z3c.autoinclude includes
by adding ``zope-conf-additional = %import sauna.reload``
into your buildout's part with *plone.recipe.zope2instance*-recipe::

  [instance]
  recipe = plone.recipe.zope2instance
  zope-conf-additional = %import sauna.reload

"""
import sys
import os

from sauna.reload.forkloop import ForkLoop
from sauna.reload.reloadpaths import ReloadPaths


reload_paths = ReloadPaths([
    os.path.join(os.getcwd(), p)
    for p in os.environ.get("RELOAD_PATH", "").split(":")
    if p
])

forkloop = ForkLoop()
forkloop.startBootTimer()

# Hook into PEP 302 laoder
from sauna.reload.monkeypatcher import MonkeyPatchingLoader
__loader__ = MonkeyPatchingLoader(sys.modules[__name__])

# Prepopulate platform.uname, before it gets lost in the stack
import platform
try:
    uname = platform.uname()
except IOError:
    # platform.uname() may throw IOError when called from a forked process
    uname = ""



from setuptools import setup, find_packages


setup(
    name="sauna.reload",
    version='0.5.5',
    description="Instant code reloading for Plone using a fork loop",
    long_description=(open("README.rst").read() + "\n" +
                      open("CHANGES.txt").read()),
    # Get more strings from
    # http://pypi.python.org/pypi?:action=list_classifiers
    classifiers=[
        "Framework :: Plone",
        "Programming Language :: Python",
    ],
    keywords="plone reload zope2 restart developer",
    author="Asko Soukka",
    author_email="asko.soukka@iki.fi",
    url="http://github.com/collective/sauna.reload",
    license="ZPL",
    packages=find_packages(exclude=["ez_setup"]),
    namespace_packages=["sauna"],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        "setuptools",
        "z3c.autoinclude",
        "watchdog>=0.6.0",
    ],
    entry_points="""
    # -*- Entry points: -*-
    [z3c.autoinclude.plugin]
    target = plone
    """,
)



# -*- coding: utf-8 -*-
# Copyright (c) 2011 University of Jyväskylä and Contributors.
#
# All Rights Reserved.
#
# Authors:
#     Esa-Matti Suuronen <esa-matti@suuronen.org>
#     Asko Soukka <asko.soukka@iki.fi>
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS

"""Overview"""

import os

from zope.publisher.browser import BrowserView

from sauna.reload import autoinclude, forkloop, reload_paths
from sauna.reload.forkloop import CannotSpawnNewChild
from sauna.reload.utils import logger


class SaunaReload(BrowserView):

    def __call__(self):
        self.forkloop = forkloop
        if self.request.get('fork', False):
            try:
                self.forkloop.spawnNewChild()
            except CannotSpawnNewChild as e:
                logger.error(str(e.args[0]))
        return self.index()

    def getConfigurationContext(self):
        try:
            from Zope2.App.zcml import _context as configuration_context
            configuration_context  # pyflakes
        except ImportError:
            from Products.Five.zcml import _context as configuration_context
        return configuration_context

    def getSaunaReloadPath(self):
        import sauna.reload
        return os.path.dirname(sauna.reload.__file__)

    def getDeferredZCMLs(self):
        cfgContext = self.getConfigurationContext()
        sauna_reload = self.getSaunaReloadPath()

        values = []
        cwd = os.getcwd() + os.path.sep
        for zcml in getattr(cfgContext, '_seen_files', ()):
            if zcml in reload_paths\
                and zcml not in autoinclude.FAILED_TO_DEFER:
                if not zcml.startswith(sauna_reload):
                    values.append(zcml.replace(cwd, ''))
        return values

    def getChildPid(self):
        return os.getpid()



# -*- coding: utf-8 -*-
# Copyright (c) 2011 University of Jyväskylä and Contributors.
#
# All Rights Reserved.
#
# Authors:
#     Esa-Matti Suuronen <esa-matti@suuronen.org>
#     Asko Soukka <asko.soukka@iki.fi>
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS

"""Default Event Implementations"""

from zope.interface import implements

from sauna.reload.interfaces import INewChildForked, INewChildIsReady


class NewChildForked(object):
    implements(INewChildForked)

    def __init__(self, forkloop):
        self.forkloop = forkloop


class NewChildIsReady(object):
    implements(INewChildIsReady)

    def __init__(self, forkloop):
        self.forkloop = forkloop
